@@ -12,6 +12,8 @@
 #ifndef UTIL_LINUX_LIST_H
 #define UTIL_LINUX_LIST_H
 
+#include "c.h"
+
 /* TODO: use AC_C_INLINE */
 #ifdef __GNUC__
 #define _INLINE_ static __inline__
@@ -33,11 +35,6 @@ struct list_head {
 	struct list_head *next, *prev;
 };
 
-#define LIST_HEAD_INIT(name) { &(name), &(name) }
-
-#define LIST_HEAD(name) \
-	struct list_head name = LIST_HEAD_INIT(name)
-
 #define INIT_LIST_HEAD(ptr) do { \
 	(ptr)->next = (ptr); (ptr)->prev = (ptr); \
 } while (0)
@@ -139,6 +136,16 @@ _INLINE_ int list_entry_is_last(struct list_head *entry, struct list_head *head)
 	return head->prev == entry;
 }
 
+/**
+ * list_entry_is_first - tests whether is entry first in the list
+ * @entry:	the entry to test.
+ * @head:	the list to test.
+ */
+_INLINE_ int list_entry_is_first(struct list_head *entry, struct list_head *head)
+{
+	return head->next == entry;
+}
+
 /**
  * list_splice - join two lists
  * @list:	the new list to add.
@@ -166,10 +173,7 @@ _INLINE_ void list_splice(struct list_head *list, struct list_head *head)
  * @type:	the type of the struct this is embedded in.
  * @member:	the name of the list_struct within the struct.
  */
-#define list_entry(ptr, type, member) ({              \
-	const typeof( ((type *)0)->member ) *__mptr = (ptr);   \
-	(type *)( (char *)__mptr - offsetof(type,member) );})
-
+#define list_entry(ptr, type, member)	container_of(ptr, type, member)
 
 #define list_first_entry(head, type, member) \
 	((head) && (head)->next != (head) ? list_entry((head)->next, type, member) : NULL)
@@ -204,6 +208,36 @@ _INLINE_ void list_splice(struct list_head *list, struct list_head *head)
 	for (pos = (head)->next, pnext = pos->next; pos != (head); \
 	     pos = pnext, pnext = pos->next)
 
+/**
+ * list_free - remove all entries from list and call freefunc()
+ *             for each entry
+ * @head:       the head for your list
+ * @type:       the type of the struct this is embedded in.
+ * @member:     the name of the list_struct within the struct.
+ * @freefunc:   the list entry deallocator
+ */
+#define list_free(head, type, member, freefunc)				\
+	do {								\
+		struct list_head *__p, *__pnext;			\
+									\
+		list_for_each_safe (__p, __pnext, (head)) {		\
+			type *__elt = list_entry(__p, type, member);	\
+			list_del(__p);					\
+			freefunc(__elt);			\
+		}							\
+	} while (0)
+
+_INLINE_ size_t list_count_entries(struct list_head *head)
+{
+	struct list_head *pos;
+	size_t ct = 0;
+
+	list_for_each(pos, head)
+		ct++;
+
+	return ct;
+}
+
 #define MAX_LIST_LENGTH_BITS 20
 
 /*
@@ -212,14 +246,16 @@ _INLINE_ void list_splice(struct list_head *list, struct list_head *head)
  * sentinel head node, "prev" links not maintained.
  */
 _INLINE_ struct list_head *merge(int (*cmp)(struct list_head *a,
-					  struct list_head *b),
+					  struct list_head *b,
+					  void *data),
+			       void *data,
 			       struct list_head *a, struct list_head *b)
 {
 	struct list_head head, *tail = &head;
 
 	while (a && b) {
 		/* if equal, take 'a' -- important for sort stability */
-		if ((*cmp)(a, b) <= 0) {
+		if ((*cmp)(a, b, data) <= 0) {
 			tail->next = a;
 			a = a->next;
 		} else {
@@ -240,7 +276,9 @@ _INLINE_ struct list_head *merge(int (*cmp)(struct list_head *a,
  * throughout.
  */
 _INLINE_ void merge_and_restore_back_links(int (*cmp)(struct list_head *a,
-						    struct list_head *b),
+						    struct list_head *b,
+						    void *data),
+					 void *data,
 					 struct list_head *head,
 					 struct list_head *a, struct list_head *b)
 {
@@ -248,7 +286,7 @@ _INLINE_ void merge_and_restore_back_links(int (*cmp)(struct list_head *a,
 
 	while (a && b) {
 		/* if equal, take 'a' -- important for sort stability */
-		if ((*cmp)(a, b) <= 0) {
+		if ((*cmp)(a, b, data) <= 0) {
 			tail->next = a;
 			a->prev = tail;
 			a = a->next;
@@ -268,7 +306,7 @@ _INLINE_ void merge_and_restore_back_links(int (*cmp)(struct list_head *a,
 		 * element comparison is needed, so the client's cmp()
 		 * routine can invoke cond_resched() periodically.
 		 */
-		(*cmp)(tail->next, tail->next);
+		(*cmp)(tail->next, tail->next, data);
 
 		tail->next->prev = tail;
 		tail = tail->next;
@@ -294,7 +332,9 @@ _INLINE_ void merge_and_restore_back_links(int (*cmp)(struct list_head *a,
  */
 _INLINE_ void list_sort(struct list_head *head,
 			int (*cmp)(struct list_head *a,
-				   struct list_head *b))
+				   struct list_head *b,
+				   void *data),
+			void *data)
 {
 	struct list_head *part[MAX_LIST_LENGTH_BITS+1]; /* sorted partial lists
 							   -- last slot is a sentinel */
@@ -316,7 +356,7 @@ _INLINE_ void list_sort(struct list_head *head,
 		cur->next = NULL;
 
 		for (lev = 0; part[lev]; lev++) {
-			cur = merge(cmp, part[lev], cur);
+			cur = merge(cmp, data, part[lev], cur);
 			part[lev] = NULL;
 		}
 		if (lev > max_lev) {
@@ -330,9 +370,9 @@ _INLINE_ void list_sort(struct list_head *head,
 
 	for (lev = 0; lev < max_lev; lev++)
 		if (part[lev])
-			list = merge(cmp, part[lev], list);
+			list = merge(cmp, data, part[lev], list);
 
-	merge_and_restore_back_links(cmp, head, part[max_lev], list);
+	merge_and_restore_back_links(cmp, data, head, part[max_lev], list);
 }
 
 #undef _INLINE_
